from niaaml.utilities import Factory
from niaaml import fitness

__all__ = [
    'FitnessFactory'
]

class FitnessFactory(Factory):
    r"""Class with string mappings to fitness class.

    Attributes:
        _entities (Dict[str, Fitness]): Mapping from strings to fitness classes.

    See Also:
        * :class:`niaaml.utilities.Factory`
    """

    def _set_parameters(self, **kwargs):
        r"""Set the parameters/arguments of the factory.
        """
        self._entities = {
            'Accuracy': fitness.Accuracy,
            'Precision': fitness.Precision,
            'CohenKappa': fitness.CohenKappa,
            'F1': fitness.F1
        }



from niaaml.data.data_reader import DataReader
from niaaml.data.basic_data_reader import BasicDataReader
from niaaml.data.csv_data_reader import CSVDataReader

__all__ = [
    'DataReader',
    'CSVDataReader',
    'BasicDataReader'
]


from unittest import TestCase
from niaaml import PipelineOptimizer, Pipeline
from niaaml.classifiers import AdaBoost, Bagging
from niaaml.preprocessing.feature_selection import SelectKBest, SelectPercentile
from niaaml.preprocessing.feature_transform import Normalizer, StandardScaler
from niaaml.data import CSVDataReader
import os
import numpy

class PipelineOptimizerTestCase(TestCase):

    def setUp(self):
        self.__data_reader = CSVDataReader(src=os.path.dirname(os.path.abspath(__file__)) + '/tests_files/dataset_header_classes.csv', has_header=True, contains_classes=True)
        self.__pipeline_optimizer = PipelineOptimizer(
            data=self.__data_reader,
            feature_selection_algorithms=['SelectKBest', 'SelectPercentile'],
            feature_transform_algorithms=['Normalizer', 'StandardScaler'],
            classifiers=['AdaBoost', 'Bagging']
        )

    def test_pipeline_optimizeer_run_works_fine(self):
        pipeline = self.__pipeline_optimizer.run('Accuracy', 10, 10, 10, 10, 'ParticleSwarmAlgorithm')
        self.assertIsInstance(pipeline, Pipeline)
        self.assertTrue(isinstance(pipeline.get_classifier(), AdaBoost) or isinstance(pipeline.get_classifier(), Bagging))
        self.assertTrue(isinstance(pipeline.get_feature_selection_algorithm(), SelectKBest) or isinstance(pipeline.get_feature_selection_algorithm(), SelectPercentile))
        self.assertTrue(pipeline.get_feature_transform_algorithm() is None or isinstance(pipeline.get_feature_transform_algorithm(), Normalizer) or isinstance(pipeline.get_feature_transform_algorithm(), StandardScaler))

    def test_pipeline_optimizer_getters_work_fine(self):
        fsas = self.__pipeline_optimizer.get_feature_selection_algorithms()
        ftas = self.__pipeline_optimizer.get_feature_transform_algorithms()
        classifiers = self.__pipeline_optimizer.get_classifiers()

        self.assertEqual(self.__pipeline_optimizer.get_data(), self.__data_reader)
        self.assertTrue((numpy.array(['AdaBoost', 'Bagging']) == numpy.array(classifiers)).all())
        self.assertTrue((numpy.array(['SelectKBest', 'SelectPercentile']) == numpy.array(fsas)).all())

        self.assertTrue((numpy.array([None, 'Normalizer', 'StandardScaler'] == numpy.array(ftas))).all())


from niaaml.preprocessing.feature_selection.feature_selection_algorithm import FeatureSelectionAlgorithm
from niaaml.preprocessing.feature_selection.variance_threshold import VarianceThreshold
from niaaml.preprocessing.feature_selection.jDEFSTH import jDEFSTH
from niaaml.preprocessing.feature_selection.select_percentile import SelectPercentile
from niaaml.preprocessing.feature_selection.select_k_best import SelectKBest
from niaaml.preprocessing.feature_selection.particle_swarm_optimization import ParticleSwarmOptimization
from niaaml.preprocessing.feature_selection.bat_algorithm import BatAlgorithm
from niaaml.preprocessing.feature_selection.differential_evolution import DifferentialEvolution
from niaaml.preprocessing.feature_selection.grey_wolf_optimizer import GreyWolfOptimizer
from niaaml.preprocessing.feature_selection.utility import FeatureSelectionAlgorithmFactory

__all__ = [
    'FeatureSelectionAlgorithm',
    'VarianceThreshold',
    'jDEFSTH',
    'SelectPercentile',
    'ParticleSwarmOptimization',
    'BatAlgorithm',
    'DifferentialEvolution',
    'GreyWolfOptimizer',
    'SelectKBest',
    'FeatureSelectionAlgorithmFactory'
]



from niaaml.utilities import Factory
import niaaml.preprocessing.feature_transform as ft

__all__ = [
    'FeatureTransformAlgorithmFactory'
]

class FeatureTransformAlgorithmFactory(Factory):
    r"""Class with string mappings to feature transform algorithms.

    Attributes:
        _entities (Dict[str, FeatureTransformAlgorithm]): Mapping from strings to feature transform algorithms.
    """

    def _set_parameters(self, **kwargs):
        r"""Set the parameters/arguments of the factory.
        """
        self._entities = {
            'Normalizer': ft.Normalizer,
            'StandardScaler': ft.StandardScaler
        }



from unittest import TestCase
from niaaml.fitness import FitnessFactory, FitnessFunction

class FitnessFactoryTestCase(TestCase):
    def setUp(self):
        self.__factory = FitnessFactory()
    
    def test_get_result_works_fine(self):
        for entry in self.__factory._entities:
            instance = self.__factory.get_result(entry)
            self.assertIsNotNone(instance)
            self.assertIsInstance(instance, FitnessFunction)
        
        with self.assertRaises(TypeError):
            self.__factory.get_result('non_existent_name')


from niaaml.classifiers.classifier import Classifier
from niaaml.utilities import MinMax
from niaaml.utilities import ParameterDefinition
from sklearn.ensemble import AdaBoostClassifier
import numpy as np

__all__ = ['AdaBoost']

class AdaBoost(Classifier):
    r"""Implementation of AdaBoost classifier.
    
    Date:
        2020

    Author
        Luka Pečnik

    License:
        MIT

    See Also:
        * :class:`niaaml.classifiers.Classifier`
    """
    Name = 'AdaBoost'

    def __init__(self, **kwargs):
        r"""Initialize AdaBoost instance.
        """
        self._params = dict(
            n_estimators = ParameterDefinition(MinMax(min=10, max=111), np.uint),
            algorithm = ParameterDefinition(['SAMME', 'SAMME.R'])
        )
        self.__ada_boost = AdaBoostClassifier()

    def set_parameters(self, **kwargs):
        r"""Set the parameters/arguments of the algorithm.
        """
        self.__ada_boost.set_params(**kwargs)

    def fit(self, x, y, **kwargs):
        r"""Fit AdaBoost.

        Arguments:
            x (numpy.ndarray[float]): n samples to classify.
            y (Iterable[any]): n classes of the samples in the x array.
        """
        self.__ada_boost.fit(x, y)

    def predict(self, x, **kwargs):
        r"""Predict class for each sample (row) in x.

        Arguments:
            x (numpy.ndarray[float]): n samples to classify.

        Returns:
            Iterable[any]: n predicted classes.
        """
        return self.__ada_boost.predict(x)

    def to_string(self):
        r"""User friendly representation of the object.

        Returns:
            str: User friendly representation of the object.
        """
        return Classifier.to_string(self).format(name=self.Name, args=self._parameters_to_string(self.__ada_boost.get_params()))


import numpy as np
from niaaml.pipeline import Pipeline
from niaaml.classifiers import ClassifierFactory
from niaaml.preprocessing.feature_selection import FeatureSelectionAlgorithmFactory
from niaaml.preprocessing.feature_transform import FeatureTransformAlgorithmFactory
from NiaPy.task import StoppingTask
from NiaPy.benchmarks import Benchmark
from NiaPy.algorithms.utility import AlgorithmUtility
from niaaml.utilities import get_bin_index

__all__ = [
    'PipelineOptimizer'
]

class PipelineOptimizer:
    r"""Optimization task that finds the best classification pipeline according to the given input.
    
    Date:
        2020

    Author
        Luka Pečnik

    License:
        MIT

    Attributes:
        __data (DataReader): Instance of any DataReader implementation.
        __feature_selection_algorithms (Optional[Iterable[str]]): Array of names of possible feature selection algorithms.
        __feature_transform_algorithms (Optional[Iterable[str]]): Array of names of possible feature transform algorithms.
        __classifiers (Iterable[Classifier]): Array of names of possible classifiers.

        __niapy_algorithm_utility (AlgorithmUtility): Utility class used to get an optimization algorithm.
    """

    def __init__(self, **kwargs):
        r"""Initialize task.
        """
        self.__data = None
        self.__feature_selection_algorithms = None
        self.__feature_transform_algorithms = None
        self.__classifiers = None
        self.__niapy_algorithm_utility = AlgorithmUtility()

        self._set_parameters(**kwargs)
    
    def _set_parameters(self, data, classifiers, feature_selection_algorithms = None, feature_transform_algorithms = None, **kwargs):
        r"""Set the parameters/arguments of the task.

        Arguments:
            data (DataReader): Instance of any DataReader implementation.
            feature_selection_algorithms (Optional[Iterable[str]]): Array of names of possible feature selection algorithms.
            feature_transform_algorithms (Optional[Iterable[str]]): Array of names of possible feature transform algorithms.
            classifiers (Iterable[Classifier]): Array of names of possible classifiers.
        """
        self.__data = data

        self.__feature_transform_algorithms = feature_transform_algorithms
        if self.__feature_transform_algorithms is not None:
            try:
                self.__feature_transform_algorithms.index(None)
            except:
                self.__feature_transform_algorithms.insert(0, None)

        self.__classifiers = classifiers
        self.__feature_selection_algorithms = feature_selection_algorithms
    
    def get_data(self):
        r"""Get data.

        Returns:
            DataReader: Instance of DataReader object.
        """
        return self.__data

    def get_feature_selection_algorithms(self):
        r"""Get feature selection algorithms.

        Returns:
            Iterable[str]: Feature selection algorithm names or None.
        """
        return self.__feature_selection_algorithms

    def get_feature_transform_algorithms(self):
        r"""Get feature transform algorithms.

        Returns:
            Iterable[str]: Feature transform algorithm names or None.
        """
        return self.__feature_transform_algorithms

    def get_classifiers(self):
        r"""Get classifiers.

        Returns:
            Iterable[str]: Classifier names.
        """
        return self.__classifiers

    def run(self, fitness_name, pipeline_population_size, inner_population_size, number_of_pipeline_evaluations, number_of_inner_evaluations, optimization_algorithm, inner_optimization_algorithm = None):
        r"""Run classification pipeline optimization process.

        Arguments:
            fitness_name (str): Name of the fitness class to use as a function.
            pipeline_population_size (uint): Number of pipeline individuals in the optimization process.
            inner_population_size (uint): Number of individuals in the hiperparameter optimization process.
            number_of_pipeline_evaluations (uint): Number of maximum evaluations.
            number_of_inner_evaluations (uint): Number of maximum inner evaluations.
            optimization_algorithm (str): Name of the optimization algorithm to use.
            inner_optimization_algorithm (Optional[str]): Name of the inner optimization algorithm to use. Defaults to the optimization_algorithm argument.
        
        Returns:
            Pipeline: Best pipeline found in the optimization process.
        """

        algo = self.__niapy_algorithm_utility.get_algorithm(optimization_algorithm)
        algo.NP = pipeline_population_size

        benchmark = _PipelineOptimizerBenchmark(self, fitness_name, inner_population_size, number_of_inner_evaluations, inner_optimization_algorithm if inner_optimization_algorithm is not None else optimization_algorithm)
        task = StoppingTask(
            D=3,
            nFES=number_of_pipeline_evaluations,
            benchmark=benchmark
            )
        algo.run(task)
        
        return benchmark.get_pipeline()

class _PipelineOptimizerBenchmark(Benchmark):
    r"""NiaPy Benchmark class implementation.
    
    Date:
        2020

    Author
        Luka Pečnik

    Attributes:
        __parent (PipelineOptimizer): Parent instance of PipelineOptimizer.
        __inner_population_size (uint): Number of individuals in the hiperparameter optimization process.
        __number_of_inner_evaluations (uint): Number of maximum inner evaluations.
        __optimization_algorithm (str): Name of the optimization algorithm to use.
        __current_best_fitness (float): Current best fitness of the optimization process.
        __current_best_pipeline (Pipeline): Current best pipeline of the optimization process.
        __fitness_name (str): Name of the fitness class to use as a function.

        __classifier_factory (ClassifierFactory): Factory for classifiers.
        __feature_transform_algorithm_factory (FeatureTransformAlgorithmFactory): Factory for feature transform algorithms.
        __feature_selection_algorithm_factory (FeatureSelectionAlgorithmFactory): Factory for feature selection algorithms.
    """
    __classifier_factory = ClassifierFactory()
    __feature_transform_algorithm_factory = FeatureTransformAlgorithmFactory()
    __feature_selection_algorithm_factory = FeatureSelectionAlgorithmFactory()

    def __init__(self, parent, fitness_name, inner_population_size, number_of_inner_evaluations, inner_optimization_algorithm):
        r"""Initialize pipeline optimizer benchmark.

        Arguments:
            parent (PipelineOptimizer): Parent instance of PipelineOptimizer.
            fitness_name (str): Name of the fitness class to use as a function.
            inner_population_size (uint): Number of individuals in the hiperparameter optimization process.
            number_of_inner_evaluations (uint): Number of maximum inner evaluations.
            inner_optimization_algorithm (str): Name of the optimization algorithm to use.
        """
        self.__parent = parent
        self.__inner_population_size = inner_population_size
        self.__number_of_inner_evaluations = number_of_inner_evaluations
        self.__optimization_algorithm = inner_optimization_algorithm
        self.__current_best_fitness = float('inf')
        self.__current_best_pipeline = None
        self.__fitness_name = fitness_name
        Benchmark.__init__(self, 0.0, 1.0)

    def __float_to_instance(self, value, collection, factory):
        r"""Get instance of object from collection using factory.

        Arguments:
            value (float): Value to map.
            collection (Iterable[str]): Array of names of possible feature selection algorithms.
            factory (Factory): Implementation of the Factory class.
        
        Returns:
            PipelineComponent: New PipelineComponent instance.
        """
        bin_index = get_bin_index(value, len(collection))

        name = collection[bin_index]
        return factory.get_result(name) if name is not None else None
    
    def get_pipeline(self):
        r"""Get best pipeline found.

        Returns:
            Pipeline: Best pipeline found.
        """
        return self.__current_best_pipeline
    
    def function(self):
        r"""Override Benchmark function.

        Returns:
            Callable[[int, numpy.ndarray[float]], float]: Fitness evaluation function.
        """
        def evaluate(D, sol):
            r"""Evaluate pipeline.

            Arguments:
                D (uint): Number of dimensionas.
                sol (numpy.ndarray[float]): Individual of population/ possible solution.
            
            Returns:
                float: Fitness.
            """
            data = self.__parent.get_data()
            pipeline = Pipeline(
                feature_selection_algorithm=self.__float_to_instance(sol[0], self.__parent.get_feature_selection_algorithms(), self.__feature_selection_algorithm_factory) if self.__parent.get_feature_selection_algorithms() is not None and len(self.__parent.get_feature_selection_algorithms()) > 0 else None,
                feature_transform_algorithm=self.__float_to_instance(sol[1], self.__parent.get_feature_transform_algorithms(), self.__feature_transform_algorithm_factory) if self.__parent.get_feature_transform_algorithms() is not None and len(self.__parent.get_feature_transform_algorithms()) > 0 else None,
                classifier=self.__float_to_instance(sol[2], self.__parent.get_classifiers(), self.__classifier_factory)
            )

            fitness = pipeline.optimize(data.get_x(), data.get_y(), self.__inner_population_size, self.__number_of_inner_evaluations, self.__optimization_algorithm, self.__fitness_name)
            if fitness < self.__current_best_fitness:
                self.__current_best_fitness = fitness
                self.__current_best_pipeline = pipeline

            return fitness
        
        return evaluate


from niaaml.utilities import Factory
import niaaml.preprocessing.feature_selection as fs
from NiaPy.benchmarks import Benchmark

__all__ = [
    'FeatureSelectionAlgorithmFactory',
    '_FeatureSelectionThresholdBenchmark'
]

class FeatureSelectionAlgorithmFactory(Factory):
    r"""Class with string mappings to feature selection algorithms.

    Attributes:
        _entities (Dict[str, FeatureSelectionAlgorithm]): Mapping from strings to feature selection algorithms.
    
    See Also:
        * :class:`niaaml.utilities.Factory`
    """

    def _set_parameters(self, **kwargs):
        r"""Set the parameters/arguments of the factory.
        """
        self._entities = {
            'jDEFSTH': fs.jDEFSTH,
            'SelectKBest': fs.SelectKBest,
            'SelectPercentile': fs.SelectPercentile,
            'VarianceThreshold': fs.VarianceThreshold,
            'BatAlgorithm': fs.BatAlgorithm,
            'DifferentialEvolution': fs.DifferentialEvolution,
            'GreyWolfOptimizer': fs.GreyWolfOptimizer,
            'ParticleSwarmOptimization': fs.ParticleSwarmOptimization
        }

class _FeatureSelectionThresholdBenchmark(Benchmark):
    r"""NiaPy Benchmark class implementation.

    Attributes:
        __best_fitness (float): Current best fitness of the optimization process.
        __best_solution (numpy.ndarray[float]): Current best solution of the optimization process.
    """
    
    def __init__(self, X, y):
        r"""Initialize feature selection benchmark.

        Arguments:
            X (numpy.ndarray[float]): Features.
            y (Iterable[any]) Expected classifier results.
        """
        self.__best_fitness = float('inf')
        self.__best_solution = None
        Benchmark.__init__(self, 0.0, 1.0)
        self.train_X, self.test_X, self.train_y, self.test_y = train_test_split(X, y, test_size=0.2)

    def get_best_solution(self):
        r"""Get best solution found.

        Returns:
            numpy.ndarray[float]: Best solution found.
        """
        return self.__best_solution

    def function(self):
        r"""Override Benchmark function.

        Returns:
            Callable[[int, numpy.ndarray[float]], float]: Fitness evaluation function.
        """

        def evaluate(D, sol):
            r"""Evaluate features.

            Arguments:
                D (uint): Number of dimensions.
                sol (numpy.ndarray[float]): Individual of population/ possible solution.
            
            Returns:
                float: Fitness.
            """
            selected = []  #array for selected features
            self.Threshold = sol[D-1]  # current threshold
            
            # select features
            for i in range(sol.shape[0] - 1):
                if sol[i] < self.Threshold:
                    pass
                else:
                    selected.append(i)

            # in the case if threshold is too low (no features selected)
            if len(selected) == 0:
                return 1
            
            lr = LogisticRegression(solver='lbfgs', max_iter=10000).fit(
                self.train_X[:, selected], self.train_y)
            accuracy = lr.score(self.test_X[:, selected], self.test_y)
            fitness = 1.0 - accuracy
            
            if fitness < self.__best_fitness:
                self.__best_fitness = fitness
                self.__best_solution = sol
            return fitness

        return evaluate


from niaaml.preprocessing.preprocessing_algorithm import PreprocessingAlgorithm

__all__ = [
    'FeatureTransformAlgorithm'
]

class FeatureTransformAlgorithm(PreprocessingAlgorithm):
    r"""Class for implementing feature transform algorithms.
    
    Date:
        2020

    Author
        Luka Pečnik

    License:
        MIT

    See Also:
        * :class:`niaaml.preprocessing.preprocessing_algorithm.PreprocessingAlgorithm`
    """

    def fit(self, x, **kwargs):
        r"""Fit implemented classifier.

        Arguments:
            x (numpy.ndarray[float]): n samples to fit transformation algorithm.
        """
        return
    
    def transform(self, x, **kwargs):
        r"""Transforms the given x data.

        Arguments:
            x (numpy.ndarray[float]): Data to transform.

        Returns:
            numpy.ndarray[float]: Transformed data.
        """
        return x



from unittest import TestCase
from niaaml.data import CSVDataReader
import os
import numpy
import csv

class CSVDataReaderTestCase(TestCase):
    def __read_csv(self, src):
        arr = []
        with open(src) as f:
            reader = csv.reader(f)
            for row in reader:
                arr1 = []
                for el in row:
                    try:
                        arr1.append(float(el))
                    except ValueError:
                        arr1.append(el)
                arr.append(arr1)
        return numpy.array(arr, dtype=object)
    
    def test_header_classes_works_fine(self):
        data_reader = CSVDataReader(src=os.path.dirname(os.path.abspath(__file__)) + '/tests_files/dataset_header_classes.csv', has_header=True, contains_classes=True)
        x = data_reader.get_x()
        y = data_reader.get_y()
        self.assertEqual(x.shape, (100, 6))
        self.assertEqual(y.shape, (100, ))

        csv_content = self.__read_csv(os.path.dirname(os.path.abspath(__file__)) + '/tests_files/dataset_header_classes.csv')
        data_reader_content = numpy.concatenate((numpy.array(x, dtype=object), numpy.array(y.reshape((100,1)), dtype=object)), axis=1)
        self.assertTrue(numpy.all(csv_content[1:] == data_reader_content))

    def test_no_header_classes_works_fine(self):
        data_reader = CSVDataReader(src=os.path.dirname(os.path.abspath(__file__)) + '/tests_files/dataset_no_header_classes.csv', has_header=False, contains_classes=True)
        x = data_reader.get_x()
        y = data_reader.get_y()
        self.assertEqual(x.shape, (100, 6))
        self.assertEqual(y.shape, (100, ))

        csv_content = self.__read_csv(os.path.dirname(os.path.abspath(__file__)) + '/tests_files/dataset_no_header_classes.csv')
        data_reader_content = numpy.concatenate((numpy.array(x, dtype=object), numpy.array(y.reshape((100,1)), dtype=object)), axis=1)
        self.assertTrue(numpy.all(csv_content == data_reader_content))

    def test_no_header_no_classes_works_fine(self):
        data_reader = CSVDataReader(src=os.path.dirname(os.path.abspath(__file__)) + '/tests_files/dataset_no_header_no_classes.csv', has_header=False, contains_classes=False)
        x = data_reader.get_x()
        y = data_reader.get_y()
        self.assertEqual(x.shape, (100, 6))
        self.assertIsNone(y)

        csv_content = self.__read_csv(os.path.dirname(os.path.abspath(__file__)) + '/tests_files/dataset_no_header_no_classes.csv')
        self.assertTrue(numpy.all(csv_content == numpy.array(x, dtype=object)))

    def test_header_no_classes_works_fine(self):
        data_reader = CSVDataReader(src=os.path.dirname(os.path.abspath(__file__)) + '/tests_files/dataset_header_no_classes.csv', has_header=True, contains_classes=False)
        x = data_reader.get_x()
        y = data_reader.get_y()
        self.assertEqual(x.shape, (100, 6))
        self.assertIsNone(y)

        csv_content = self.__read_csv(os.path.dirname(os.path.abspath(__file__)) + '/tests_files/dataset_header_no_classes.csv')
        self.assertTrue(numpy.all(csv_content[1:] == numpy.array(x, dtype=object)))


from niaaml.preprocessing import feature_selection
from niaaml.preprocessing import feature_transform
from niaaml.preprocessing.preprocessing_algorithm import PreprocessingAlgorithm

__all__ = [
    'feature_selection',
    'feature_transform',
    'PreprocessingAlgorithm'
]